/*
 * Copyright (c) 2012-2014,2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Fri Mar 23 2012
 */
#include "acgenomix.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <dirent.h>
#include <dlfcn.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "genomix.h"


/* --- global data --------------------------------------------------------- */

/* client handles */
struct gnmx_handle *gnmx_clients = NULL;


/* --- local data ---------------------------------------------------------- */

/* rpaths */
static const char **rpaths = NULL;
static size_t nrpaths = 0;

static void	gnmx_requestcb(genom_client c, int rqstid, int done,
			genom_event report, void *output, const void *exdetail,
			void *data);
static int	gnmx_module_cmd(ClientData data, Tcl_Interp *interp,
			int objc, Tcl_Obj *const objv[]);
static void	gnmx_delete_module_cmd(ClientData data);

void gnmx_event_proc(ClientData data, int mask);

static void
gnmx_strfree(char *s)
{
  free(s);
}

/* --- gnmx_rpath ---------------------------------------------------------- */

int
gnmx_rpath(ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
{
  const char *path;
  const char **p;
  size_t i;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "path");
    return TCL_ERROR;
  }

  path = Tcl_GetString(objv[1]);
  for(i=0; i<nrpaths; i++)
    if (!strcmp(rpaths[i], path)) return TCL_OK;

  p = realloc(rpaths, (nrpaths+1)*sizeof(*rpaths));
  if (!p) {
    Tcl_AppendResult(interp, strerror(errno), NULL);
    return TCL_ERROR;
  }
  rpaths = p;
  rpaths[nrpaths] = strdup(path);
  nrpaths++;

  Tcl_SetObjResult(interp, objv[1]);
  return TCL_OK;
}


/* --- gnmx_load ----------------------------------------------------------- */

static void *
gnmx_dlopen(Tcl_Interp *interp, const char *path)
{
  Tcl_SavedResult r;
  char msg[256];
  void *h;

  h = dlopen(path, RTLD_NOW | RTLD_LOCAL);
  if (!h) {
    const char *emsg = dlerror();
    snprintf(msg, sizeof(msg), "infox {%s} 1", emsg);
    Tcl_SaveResult(interp, &r);
    Tcl_Eval(interp, msg);
    Tcl_RestoreResult(interp, &r);
    if (*Tcl_GetStringResult(interp))
      Tcl_AppendResult(interp, "\n", NULL);
    Tcl_AppendResult(interp, emsg, NULL);
  }

  return h;
}

static void *
gnmx_lookup(Tcl_Interp *interp, const char *module, char **loaded)
{
  static const char dotso[] = ".so";

  char path[PATH_MAX];
  struct dirent *dirent;
  struct stat sb;
  int nodotso;
  void *dlh;
  DIR *dir;
  size_t i;

  dlh = NULL;

  /* check if 'module' ends with .so */
  nodotso = strncmp(
    module + strlen(module) - strlen(dotso), dotso, strlen(dotso));

  /* try rpaths first */
  for(i = 0; i<nrpaths; i++) {
    /* try a raw file */
    snprintf(path, sizeof(path), "%s/%s", rpaths[i], module);
    if (!stat(path, &sb) && S_ISREG(sb.st_mode)) {
      dlh = gnmx_dlopen(interp, path);
      if (dlh) break;
    }

    /* try with .so appended (if not present already) */
    if (nodotso) {
      snprintf(path, sizeof(path), "%s/%s%s", rpaths[i], module, dotso);
      if (!stat(path, &sb) && S_ISREG(sb.st_mode)) {
        dlh = gnmx_dlopen(interp, path);
        if (dlh) break;
      }
    }
  }

  /* no matching file in rpaths, try system defaults */
  if (!dlh) {
    snprintf(path, sizeof(path), "%s", module);
    dlh = gnmx_dlopen(interp, path);
  }

  /* also try with .so appended (if not present already) */
  if (!dlh && nodotso) {
    snprintf(path, sizeof(path), "%s%s", module, dotso);
    dlh = gnmx_dlopen(interp, path);
  }

  /* finally try in genomix prefix at the last resort */
  if (!dlh) {
    snprintf(path, sizeof(path), "%s/genom", GENOMIX_LIBDIR);
    dir = opendir(path);
    if (dir) {
      while ((dirent = readdir(dir))) {
        if (!strcmp(dirent->d_name, ".") || !strcmp(dirent->d_name, ".."))
          continue;

        snprintf(path, sizeof(path), "%s/genom/%s/plugins/%s",
                 GENOMIX_LIBDIR, dirent->d_name, module);
        if (!stat(path, &sb) && S_ISREG(sb.st_mode))
          dlh = gnmx_dlopen(interp, path);

        /* and with .so appended (if not present already) */
        if (!dlh && nodotso) {
          snprintf(path, sizeof(path), "%s/genom/%s/plugins/%s%s",
                   GENOMIX_LIBDIR, dirent->d_name, module, dotso);
          if (!stat(path, &sb) && S_ISREG(sb.st_mode))
            dlh = gnmx_dlopen(interp, path);
        }

        if (dlh) break;
      }

      closedir(dir);
    }
  }

  /* no more hint */
  if (loaded) *loaded = dlh ? strdup(path) : NULL;
  return dlh;
}

int
gnmx_load(ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
{
  struct genom_client_info **info;
  struct gnmx_handle *h;
  const char *name;
  const char *module;
  Tcl_CmdInfo cmdinfo;
  Tcl_Obj *cmd;
  Tcl_Obj **largv;
  char msg[256];
  size_t i;
  int largc;
  void *ath;

  if (objc < 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "path args");
    return TCL_ERROR;
  }
  module = Tcl_GetString(objv[1]);

  h = malloc(sizeof(*h));
  if (!h) {
    Tcl_AppendResult(interp, strerror(errno), NULL);
    return TCL_ERROR;
  }
  h->c = NULL;

  /* lookup plugin */
  h->dlh = gnmx_lookup(interp, module, &h->path);
  if (!h->dlh) goto error;

  /* we loaded something */
  snprintf(msg, sizeof(msg), "infox {%s: loaded} 1", h->path);
  Tcl_Eval(interp, msg);
  Tcl_ResetResult(interp);

  /* retrieve entry point */
  info = dlsym(h->dlh, GENOM_CLIENT_DLSYM);
  if (!info) {
    Tcl_AppendResult(interp, dlerror(), NULL);
    goto error;
  }
  h->info = *info;
  if (h->info->protocol != genom_client_protocol) {
    char num[16];
    Tcl_AppendResult(interp, "client protocol mismatch, client ", NULL);
    snprintf(num, sizeof(num), "%d", h->info->protocol);
    Tcl_AppendResult(interp, num, " vs. genomix ", NULL);
    snprintf(num, sizeof(num), "%d", genom_client_protocol);
    Tcl_AppendResult(interp, num, NULL);
    goto error;
  }

  /* get args */
  if (Tcl_ListObjGetElements(interp, objv[2], &largc, &largv) != TCL_OK)
    goto error;
  {
    char *argv[largc];
    genom_event ex;
    const void *exdetail;

    for (i = 0; i<largc; i++)
      argv[i] = Tcl_GetString(largv[i]);

    /* setup communication - protect against bad atexit() */
    gnmx_atexit_push(h->dlh, &ath);
    h->c = h->info->init(largc, argv, &ex, &exdetail);
    if (!h->c) {
      Tcl_SetResult(interp,
                    h->info->json_error(NULL, ex, exdetail), gnmx_strfree);
      gnmx_atexit_pop(ath);
      goto error;
    }
    gnmx_atexit_pop(ath);
  }

  /* create command */
  gnmx_atexit_push(h->dlh, &ath);
  name = h->info->instance(h->c);
  gnmx_atexit_pop(ath);
  cmd = Tcl_NewStringObj("::module::", -1);
  Tcl_AppendToObj(cmd, name, -1);
  Tcl_IncrRefCount(cmd);

  if (Tcl_GetCommandInfo(interp, Tcl_GetString(cmd), &cmdinfo)) {
    /* exists: reuse existing command */
    Tcl_AppendResult(interp, name, NULL);
    gnmx_atexit_push(h->dlh, &ath);
    h->info->fini(h->c);
    gnmx_atexit_pop(ath);
    gnmx_atexit_invoke(h->dlh);
    dlclose(h->dlh);
    free(h);
    Tcl_DecrRefCount(cmd);
    return TCL_OK;
  }

  h->token = Tcl_CreateObjCommand(
    interp, Tcl_GetString(cmd), gnmx_module_cmd, h, gnmx_delete_module_cmd);
  Tcl_DecrRefCount(cmd);
  if (!h->token) goto error;

  h->interp = interp;
  gnmx_atexit_push(h->dlh, &ath);
  h->eventfd = h->info->eventfd(h->c);
  gnmx_atexit_pop(ath);
  h->next = gnmx_clients;
  gnmx_clients = h;
  gnmx_watch(h);

  Tcl_AppendResult(interp, name, NULL);
  return TCL_OK;

error:
  if (h) {
    if (h->c) {
      gnmx_atexit_push(h->dlh, &ath);
      h->info->fini(h->c);
      gnmx_atexit_pop(ath);
    }
    if (h->dlh) {
      gnmx_atexit_invoke(h->dlh);
      dlclose(h->dlh);
    }
    if (h->path) free(h->path);
    free(h);
  }
  return TCL_ERROR;
}


/* --- gnmx_usage ---------------------------------------------------------- */

int
gnmx_usage(ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
{
  const struct genom_client_option *options;
  struct genom_client_info **info;
  const char *module;
  char msg[256];
  char *path;
  void *dlh;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "path");
    return TCL_ERROR;
  }
  module = Tcl_GetString(objv[1]);

  /* lookup plugin */
  dlh = gnmx_lookup(interp, module, &path);
  if (!dlh) return TCL_ERROR;

  /* we loaded something */
  snprintf(msg, sizeof(msg), "infox {%s: found} 1", path);
  Tcl_Eval(interp, msg);
  Tcl_ResetResult(interp);

  /* retrieve entry point */
  info = dlsym(dlh, GENOM_CLIENT_DLSYM);
  if (!info) {
    Tcl_AppendResult(interp, dlerror(), NULL);
    goto error;
  }
  if ((*info)->protocol != genom_client_protocol) {
    char num[16];
    Tcl_AppendResult(interp, "client protocol mismatch, client ", NULL);
    snprintf(num, sizeof(num), "%d", (*info)->protocol);
    Tcl_AppendResult(interp, num, " vs. genomix ", NULL);
    snprintf(num, sizeof(num), "%d", genom_client_protocol);
    Tcl_AppendResult(interp, num, NULL);
    goto error;
  }

  /* build the usage string */
  Tcl_AppendResult(interp, "[", NULL);
  for(options = (*info)->options; options->option; options++) {
    Tcl_AppendResult(interp,
                     "{\"option\":\"", options->option, "\",",
                     "\"descr\":\"", options->descr, "\"}",
                     options[1].option? "," : NULL,
                     NULL);
  }
  Tcl_AppendResult(interp, "]", NULL);

  if (dlh) dlclose(dlh);
  if (path) free(path);
  return TCL_OK;

error:
  if (dlh) dlclose(dlh);
  if (path) free(path);
  return TCL_ERROR;
}


/* --- gnmx_module_cmd ----------------------------------------------------- */

static int
gnmx_module_cmd(ClientData data, Tcl_Interp *interp,
                int objc, Tcl_Obj *const objv[])
{
  enum cidx {
    cidx_path, cidx_info, cidx_send, cidx_abort, cidx_clean, cidx_read
  };
  static const char *args[] = {
    [cidx_path] = "path", [cidx_info] = "info", [cidx_send] = "send",
    [cidx_abort] = "abort", [cidx_clean] = "clean", [cidx_read] = "read", NULL
  };

  struct gnmx_handle *h = data;
  enum cidx i;
  void *ath;
  int s, e;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 0, objv, "command ?args?");
    return TCL_ERROR;
  }

  e = Tcl_GetIndexFromObj(interp, objv[1], args, "subcommand", 0, (int *)&i);
  if (e != TCL_OK) return e;

  switch(i) {
    case cidx_path: {
      Tcl_AppendResult(interp, h->path?h->path:"<null>", NULL);
      return h->path ? TCL_OK : TCL_ERROR;
    }

    case cidx_info: {
      Tcl_AppendResult(interp, "{", NULL);
      Tcl_AppendResult(interp, "\"name\":\"", h->info->name, "\",", NULL);
      Tcl_AppendResult(interp, "\"instance\":\"",
                       h->info->instance(h->c), "\",", NULL);

      Tcl_AppendResult(interp, "\"services\":{", NULL);
      for(s=0; s<h->info->nservices; s++) {
        if (s) Tcl_AppendResult(interp, ",", NULL);

        Tcl_AppendResult(interp, "\"",
                         h->info->services[s].name, "\":{", NULL);
        Tcl_AppendResult(interp, "\"input\":",
                         h->info->services[s].input, ",", NULL);
        Tcl_AppendResult(interp, "\"output\":",
                         h->info->services[s].output, ",", NULL);
        Tcl_AppendResult(interp, "\"meta\":", h->info->services[s].meta, NULL);
        Tcl_AppendResult(interp, "}", NULL);
      }
      Tcl_AppendResult(interp, "},", NULL);

      Tcl_AppendResult(interp, "\"ports\":{", NULL);
      for(s=0; s<h->info->nports; s++) {
        if (s) Tcl_AppendResult(interp, ",", NULL);

        Tcl_AppendResult(interp, "\"", h->info->ports[s].name, "\":{", NULL);
        if (h->info->ports[s].simple)
          Tcl_AppendResult(interp, "\"simple\":1,", NULL);
        else
          Tcl_AppendResult(interp, "\"simple\":0,", NULL);
        Tcl_AppendResult(interp, "\"data\":",
                         h->info->ports[s].data, ",", NULL);
        Tcl_AppendResult(interp, "\"meta\":", h->info->ports[s].meta, NULL);
        Tcl_AppendResult(interp, "}", NULL);
      }
      Tcl_AppendResult(interp, "}", NULL);

      Tcl_AppendResult(interp, "}", NULL);
      break;
    }

    case cidx_send: {
      const char *rname, *rarg;
      genom_event s;
      char *e;
      int a, r, err;

      if (objc != 4) {
        Tcl_WrongNumArgs(interp, 1, objv, "service arg");
        return TCL_ERROR;
      }
      rname = Tcl_GetString(objv[2]);
      rarg = Tcl_GetString(objv[3]);

      for(r=0; r < h->info->nservices; r++)
        if (!strcmp(h->info->services[r].name, rname)) break;
      if (r >= h->info->nservices) {
        Tcl_AppendResult(interp, "unknown service ", rname, NULL);
        return TCL_ERROR;
      }

      {
        char input[h->info->services[r].input_size];

        gnmx_atexit_push(h->dlh, &ath);
        h->info->services[r].init_input(input);
        err = h->info->services[r].json_scan(input, rarg, &e);
        if (err) h->info->services[r].fini_input(input);
        gnmx_atexit_pop(ath);
        if (err) {
          char ectx[64];
          int n;

          Tcl_AppendResult(interp, "invalid json string at offset ", NULL);
          snprintf(ectx, sizeof(ectx), "%d", (int)(e-rarg));
          Tcl_AppendResult(interp, ectx, ": ", NULL);
          n = e-rarg > 15 ? 15 : e-rarg;
          snprintf(ectx, sizeof(ectx), "%.*s<here>%.15s", n, e-n, e);
          Tcl_AppendResult(interp, ectx, NULL);
          return TCL_ERROR;
        }

        gnmx_atexit_push(h->dlh, &ath);
        s = h->info->services[r].send(h->c, input,
                                    gnmx_requestcb, gnmx_requestcb, h, &a);
        h->info->services[r].fini_input(input);
        gnmx_atexit_pop(ath);
        if (s) {
          Tcl_SetResult(interp,
                        h->info->json_error(h->c, NULL, NULL), gnmx_strfree);
          return TCL_ERROR;
        }
      }

      Tcl_SetObjResult(interp, Tcl_NewIntObj(a));
      break;
    }

    case cidx_abort:
    case cidx_clean: {
      genom_event s;
      int a;

      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "rqstid");
        return TCL_ERROR;
      }
      if (Tcl_GetIntFromObj(interp, objv[2], &a) != TCL_OK)
        return TCL_ERROR;

      gnmx_atexit_push(h->dlh, &ath);
      switch(i) {
        case cidx_abort:	s = h->info->abort(h->c, a); break;
        case cidx_clean:	s = h->info->clean(h->c, a); break;
        default:		assert(0);
      }
      gnmx_atexit_pop(ath);
      if (s) {
        Tcl_SetResult(interp,
                      h->info->json_error(h->c, NULL, NULL), gnmx_strfree);
        return TCL_ERROR;
      }
      break;
    }

    case cidx_read: {
      const char *port, *name;
      genom_event s;
      char *data;
      char *json;
      int p, err;

      if (objc != 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "port subname");
        return TCL_ERROR;
      }
      port = Tcl_GetString(objv[2]);
      name = Tcl_GetString(objv[3]);

      for(p=0; p < h->info->nports; p++)
        if (!strcmp(h->info->ports[p].name, port)) break;
      if (p >= h->info->nports) {
        Tcl_AppendResult(interp, "unknown port ", port, NULL);
        return TCL_ERROR;
      }

      data = malloc(h->info->ports[p].data_size);
      if (!data) {
        Tcl_AppendResult(interp, "not enough memory", NULL);
        return TCL_ERROR;
      }
      gnmx_atexit_push(h->dlh, &ath);
      h->info->ports[p].init_data(data);
      if (h->info->ports[p].simple)
        s = h->info->ports[p].read.simple(h->c, data);
      else {
        if (name[0] == '\0') {
          Tcl_AppendResult(interp, "missing port set name", NULL);
          return TCL_ERROR;
        }
        s = h->info->ports[p].read.multiple(h->c, name, data);
      }
      gnmx_atexit_pop(ath);
      if (s) {
        free(data);
        Tcl_SetResult(interp,
                      h->info->json_error(h->c, NULL, NULL), gnmx_strfree);
        return TCL_ERROR;
      }
      err = h->info->ports[p].json_print(&json, data);
      free(data);
      if (err) {
        Tcl_SetResult(interp, strerror(err), TCL_VOLATILE);
        return TCL_ERROR;
      }

      Tcl_SetObjResult(interp, Tcl_NewStringObj(json, -1));
      free(json);
      break;
    }
  }

  return TCL_OK;
}


/* --- gnmx_requestcb ------------------------------------------------------ */

static void
gnmx_requestcb(genom_client c, int rqstid, int done, genom_event report,
               void *output, const void *exdetail, void *data)
{
  struct gnmx_handle *h = data;
  const struct genom_service_info *r;
  static const int argc = 6;
  Tcl_Obj *argv[argc];
  char *err, *json = NULL;
  int i, s, e;

  /* command */
  argv[0] = Tcl_NewStringObj("request::event", -1);
  argv[1] = Tcl_NewStringObj(h->info->instance(c), -1);
  argv[2] = Tcl_NewIntObj(rqstid);

  /* status */
  if (report == genom_ok)
    argv[3] = Tcl_NewStringObj(done?"done":"sent", -1);
  else
    argv[3] = Tcl_NewStringObj("error", -1);

  /* error string, or empty */
  if (report) {
    err = h->info->json_error(c, report, exdetail);
    argv[4] = Tcl_NewStringObj(err, -1);
    free(err);
  } else if (output) {
    r = h->info->service_info(c, rqstid);
    if (!r)
      argv[4] = Tcl_NewStringObj("\"internal error: rqstid with no data\"", -1);
    else {
      e = r->json_print(&json, output);
      if (e) {
        argv[4] = Tcl_NewStringObj(strerror(e), -1);
      } else
        argv[4] = Tcl_NewStringObj("", -1);
    }
  } else
    argv[4] = Tcl_NewStringObj("", -1);

  /* json output */
  if (json) {
    argv[5] = Tcl_NewStringObj(json, -1);
    free(json);
  } else
    argv[5] = Tcl_NewStringObj("", -1);

  /* invoke callback */
  for(i = argc-1; i >= 0; i--) Tcl_IncrRefCount(argv[i]);
  s = Tcl_EvalObjv(h->interp, argc, argv, TCL_EVAL_GLOBAL);
  for(i = argc-1; i >= 0; i--) Tcl_DecrRefCount(argv[i]);
  if (s != TCL_OK)
    Tcl_BackgroundError(h->interp);
}


/* --- gnmx_delete_cmd ----------------------------------------------------- */

static void
gnmx_delete_module_cmd(ClientData data)
{
  struct gnmx_handle *h = data;
  struct gnmx_handle *i;
  void *ath;

  if (h) {
    if (h == gnmx_clients)
      gnmx_clients = h->next;
    else {
      for(i = gnmx_clients; i->next && i->next != h; i = i->next)
        /* empty body */;
      if (i->next == h) i->next = h->next;
    }

    gnmx_unwatch(h);
    if (h->c){
      gnmx_atexit_push(h->dlh, &ath);
      h->info->fini(h->c);
      gnmx_atexit_pop(ath);
    }

    if (h->dlh) {
      gnmx_atexit_invoke(h->dlh);
      dlclose(h->dlh);
    }
    if (h->path) free(h->path);
    free(h);
  }
}
